#ifndef _SAORI_H_
#define _SAORI_H_


#include <vector>
#include <string>
#include <map>
#include <sstream>
#include <boost/ref.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>

#include <windows.h>


// Saoriの返り値
enum saori_result_id
{
    SAORI_200_OK,
    SAORI_210_No_Contents,
    SAORI_400_Bad_Request,
    SAORI_500_Internal_Server_Error
};


// Saoriエラー型
class saori_error
{
private:
    saori_result_id _id;
    std::string _what;
    
public:
    // コンストラクタ
    saori_error(saori_result_id id, const std::string &w)
        : _id(id), _what(w){}
    // エラーの返り値を取得
    const saori_result_id id() const
    {
        return _id;
    }
    // エラーの原因
    const std::string &what() const
    {
        return _what;
    }
};

// パーサクラス
class request_parser
{   
public:
    // Argument[n]が入るベクタ
    typedef std::vector<std::string> ArgumentType;
    
private:
    std::string _charset;
    std::string _command;
    std::string _securitylevel;
    std::string _sender;
    const std::string &_request_data;
    bool _get_ver;

protected:
    ArgumentType _args;
    int n;
    // あまり無名関数を使いまわすとコンパイラが悲鳴を上げるので救済処置
    struct set_arg_t
    {
        request_parser * const _p;
        set_arg_t(request_parser * const p) : _p(p){}
        template <typename T>
        void operator()(const T &b, const T &e) const
        {
            _p->_args.reserve(_p->n + 1);
            _p->_args[_p->n] = std::string(b,e);
        }
    };
    
    friend class set_arg_t;
    
public:
    // コンストラクタ
    request_parser(const std::string& data) : _request_data(data), _get_ver(false), _args(10) {}
    // パース開始
    void parse();
    // Argumentの取得
    const ArgumentType &arguments() const
    {
        return _args;
    }
    // 文字コードの取得
    const std::string &charset() const
    {
        return _charset;
    }
    // セキュリティーレベルの取得
    const std::string &securitylevel() const
    {
        return _securitylevel;
    }
    // sender
    const std::string &sender() const
    {
        return _sender;
    }
    // コマンド
    const std::string &command() const
    {
        return _command;
    }
    
};


// さおりインターフェースのスーパークラス
class saori_interface_base
{
public:
    // リクエスト
    virtual bool request(const std::string&, std::string&) = 0;

    // pathを入れる。
    virtual void set_path(const std::string&) = 0;
    
    // load呼び出しのときにときに呼ばれる。
    // path はpath()でとる。
    virtual void load_module() = 0;
    
    // unload呼び出しのときにときに呼ばれる。
    virtual void unload_module() = 0;
};


// Saoriインターフェースクラス
template <typename Type, typename Rule>
class saori_interface : public saori_interface_base
{
public:
    // 派生クラスの型
    typedef Type BaseType;
    // そのポインタ型
    typedef BaseType * const PtrBase;
    // Argument[n]が入るベクタ
    typedef std::vector<std::string> ArgumentType;
    // Result[n]が入る
    typedef std::vector<std::string> ResultType;
    // スロット関数の返り値の型
    typedef std::pair<std::string, ResultType> ResultPair;
    // 子クラスから渡されるスロット関数の型
    typedef boost::function<bool(PtrBase, const ArgumentType&, ResultPair&)>
        SlotFunction;
    typedef Rule RuleType;
        
private:
    // 格納されるスロット関数の型
    typedef boost::function<bool(const ArgumentType&, ResultPair&)>
        InnerSlotFunction;
    // スロット関数を格納する連想配列
    typedef std::map<std::string, InnerSlotFunction> FuncMap;

    // スロット関数の入っている連想配列
    FuncMap _funcmap;
    // loadでもらえるpath
    std::string _path;
    // 指定されたコード
    std::string _charset;
    // セキュリティー
    std::string _securitylevel;
    //senderヘッダ
    std::string _sender;
    // 
    std::string _myname;
    
protected:
    // スロットの追加
    void add_slot(const std::string &arg0name, SlotFunction func)
    {
        _funcmap.insert(std::make_pair(arg0name, boost::bind(func, (PtrBase)this, _1, _2)));
    }
    // pathの取得
    const std::string &path() const
    {
        return _path;
    }
    // 文字コードの取得
    const std::string &get_charset() const
    {
        return _charset;
    }
    // 文字コードのセット
    void set_charset(const std::string &set)
    {
        _charset = set;
    }
    // セキュリティーレベルの取得
    const std::string &securitylevel() const
    {
        return _securitylevel;
    }
    //
    const std::string &sender() const
    {
        return _sender;
    }
    
    void set_myname(const std::string &name)
    {
        _myname = name;
    }
    
public:
    // リクエストをして返り値をもらう
    bool request(const std::string &req, std::string &result);

    // pathを入れる。
    void set_path(const std::string& p)
    {
        _path = p;
    }
    
    // load呼び出しのときにときに呼ばれる。
    // path はpath()でとる。
    virtual void load_module(){}
    
    // unload呼び出しのときにときに呼ばれる。
    virtual void unload_module(){}
    

};


struct saori_con_base
{
    virtual void load(const std::string&) = 0;
    virtual void unload() = 0;
    virtual void request(const std::string&, std::string&) = 0;
};


template<typename Rule>
struct saori_con : public saori_con_base
{
    typedef Rule RuleType;
    
    void load(const std::string &path)
    {
        RuleType::load(path);
    }
    
    void unload()
    {
        RuleType::unload();
    }
    
    void request(const std::string &req, std::string &res)
    {
        RuleType::request(req, res);
    }
};



extern saori_interface_base *_saori_interface_object_for_module_;




template <typename Saori>
struct unlazylodable
{
    static void load(const std::string &path)
    {
        _saori_interface_object_for_module_ = new Saori;
        // path のセット
        _saori_interface_object_for_module_->set_path(path); 
        // ユーザ定義のロード時の処理を。
        _saori_interface_object_for_module_->load_module();
    }
    
    
    static void unload()
    {
        // unload時のユーザ定義処理。
        _saori_interface_object_for_module_->unload_module();
        delete _saori_interface_object_for_module_;
    }
    
    static void request(const std::string &req, std::string &res)
    {           
        // リクエスト処理。
        _saori_interface_object_for_module_->request(req, res);
    }
};



template <typename Saori>
struct lazylodable
{
    static void load(const std::string &path)
    {}
    
    
    static void unload()
    {}
    
    static void request(const std::string &req, std::string &res)
    {
        boost::scoped_ptr<Saori> saori_object(new Saori);           
        // リクエスト処理。
        saori_object->request(req, res);
    }
};



// リクエストをパースしてスロットにる
template <typename Type, typename Rule>
bool saori_interface<Type, Rule>::request
(const std::string &req, std::string &result)
{
    request_parser parser(req);

    try{
        // パース
        parser.parse();
        // スロット
        typename FuncMap::iterator func_pair;
        if(parser.command() == "GET Version"){
            result = "SHIORI/1.0 210 No Contents\r\n\r\n";
            return true;
        }
        _charset = parser.charset();
        _securitylevel = parser.securitylevel();
        _sender = parser.sender();
        // Argumentのないさおり呼び出しには未対応
        if(parser.arguments().size() == 0)
            throw saori_error(SAORI_400_Bad_Request, "Bad argument1.");
        // 探しにいく
        func_pair = _funcmap.find(parser.arguments()[0]);
        if(func_pair == _funcmap.end())
            throw saori_error(SAORI_400_Bad_Request, parser.arguments()[0]);
        
        // スロットの返り値
        ResultPair res_pair;
        // 呼び出し
        if(func_pair->second(boost::ref(parser.arguments()), boost::ref(res_pair))){
            // trueが返ったら 200 OK
            std::stringstream sresult;
            sresult << "SAORI/1.0 200 OK\r\n";
            sresult << "Charset: " + _charset + "\r\n";
            sresult << "Sender: " + _myname + "\r\n";
            if(res_pair.first != "")
            sresult << "Result: " + res_pair.first + "\r\n";
            typename ResultType::iterator itr;
            int n = 0;
            for(itr = res_pair.second.begin(); itr !=res_pair.second.end(); itr++)
                sresult << "Value" << n++ << ": " << *itr << "\r\n";
            sresult << "\r\n";
            result = sresult.str();
        }else{
            // falseが返ったら 210 No Contents
            result = "SHIORI/1.0 210 No Contents\r\n\r\n";
        }
        return true;
    }catch(saori_error &err){
        if(err.id() == SAORI_400_Bad_Request)
            result = "SAORI/1.0 400 Bad Request\r\n\r\n";
        else
            result = "SAORI/1.0 500 Internal Server Error\r\n\r\n";
        return false;
    }
}


saori_con_base *saori_controler_initialize();


// 一番外で
// SAORI_INTERFACE_MODULE(hogeclass)
// と書いてDLLでコンパイルする
#define SAORI_INTERFACE_MODULE(class_name) \
  saori_con_base *saori_controler_initialize() \
  { \
    return new saori_con<class_name::RuleType>; \
  }

#endif
